## Copyright (c) 2020
## Global Climate Forum e.V.
## http://www.globalclimateforum.org

## Authors:
##     Steffen Fuerst

## This file is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This file is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## version 0.11

if ("rstudioapi" %in% loadedNamespaces() && rstudioapi::isAvailable()) {
  setwd(dirname(rstudioapi::getActiveDocumentContext()$path))
}
source("platform.R")

## This file contains the model simulation of the tutorial for the GCF
## computation platform, Version 0. The tutorial itself is more
## extensive, so it's usefull to read also the corresponding pdf.

## We have two types of agents, ~BUYER~ and ~SELLER~. And we have also
## two types of networks. The first is fixed and describes the sellers
## known to a single buyer. Since the direction of an edge determines
## the flow of information and the buyer needs the price information
## to calculate his demand, it is a network from sellers to buyers. We
## call this network ~KNOWN_SELLER~.

## The second network ~BOUGHT_FROM~ will be constructed in
## the transition function called ~calcDemand~. This network
## describes from which seller a buyer bought something, so it will go
## from the buyers to the sellers. Beside ~calcDemand~ which will be the
## implementation of equation (\ref{eqn:demand}) we will also implement a transition function
## with the name ~calcPrice~, which implements equation (\ref{eqn:price}).

## So we know already the ~agentTypes~ and ~edgeTypes~, and the
## ~modelParams~ are just a tibble, which must contain the parameter
## ~numPEs~ beside the model parameters we define for
## ourselve. ~numPEs~ determines on how many processing entities the
## simulation will be distributed[fn:9]. So we have already:

agentTypes <- c("BUYER", "SELLER")
edgeTypes <- c("KNOWN_SELLERS", "BOUGHT_FROM")
modelParams <- tibble(numBuyer = 50, numSeller = 5, knownSellers = 2, numPEs = 5)

## ~stateInitFunc~

## The ~stateInitFunc~ is a function, that will get the modelParams as
## argument, and must return the inital model state, in form of a list of tables,
## with one table for each agent and edge type. We want to initialize our
## buyers with values for $\alpha$ between 0 and 1, with values for $B$
## between 1 and 100 and the sellers initial prices $p$ between 0.5
## and 1.5. Also we must create the ~KNOWN_SELLERS~ network, by selecting
## for each ~BUYER~ randomly ~modelParams$knownSellers~ seller.

## So, our implementation of the ~stateInitFunc~ is the following:

initModel <- function(modelParams) {
  initState <- function(modelParams) {
    nb <- modelParams$numBuyer
    ns <- modelParams$numSeller
    list(
        "BUYER" = tibble(
            alpha = runif(nb),
            B = runif(nb, max = 100) %>% ceiling,
            x = 0,
            y = 0
        ),
        "SELLER" = tibble(p = runif(ns, min = 0.5, max = 1.5),
                          sum_y = 0)
    )
  }

  initNetwork <- function(modelParams, modelState) {
    sampleSellers <- function() {
      sample_n(modelState$SELLER, modelParams$knownSellers) %>% pull(ID)
    }

    list("KNOWN_SELLERS" = 
           modelState$BUYER$ID %>%
           map(~ tibble(fromID = sampleSellers(), 
                        toID = .x)) %>%
           bind_rows)
  }

  agentState <- initState(modelParams) %>% withStringID 
  c(agentState, initNetwork(modelParams, agentState))
}

## ~globalInitFunc~

## For the ~initSimulation~ function we still need a ~globalInitFunc~
## and the ~globalTypes~ list. We need something that give us time
## series for the variables we want to observe. In our case, we want
## to know the trajectories of the price and demands. So we create in
## the ~globalInitFunc~ an empty tibble with the corresponding three
## columns, and name this tibble ~OBSERVER~:

initObserver <- function(modelParams) {
  list("OBSERVER" =
         tibble(
             p = numeric(),
             sum_x = numeric(),
             sum_y = numeric()
         ))
}

## ~calcDemand~

## The first transition function we are implementing calculates the
## demand for the goods $x$ and $y$.

## In the case that the agent is a ~BUYER~, one of the available
## sellers is selected using the ~sample_n~ function. Then the agent
## updates it's demand for the goods $x$ and $y$ and adds the new state
## to the ~BUYERs~ of the overall new model state. 

## The agent also adds an edge to the ~BOUGHT_FROM~ network, which is
## then used in the next transition function by the sellers to sum up
## the demand and calculate the new price.

## The ~KNOWN_SELLERS~ network is constant, so we will add 
## ~KNOWN_SELLERS~ to the list of ~invariantTypes~ when 
## ~applyTransition~ is called. As the sellers do not change their 
## state, ~SELLER~ will be also added to ~invariantTypes~.

## The ~KNOWN_SELLER~ network is constant, but must be recreated in
## this platform version by each transition function. However, this
## can easily be done by letting each agent return all edges where
## itself is at the head position. The sum of these edges will then
## yield the same network.

## A seller does nothing else but return its state. Otherwise the model
## state would be without any seller after applying this transition function.

calcDemand <- function(agent, type, modelState) {
  if (type == "BUYER") {
    seller <- modelState$SELLER %>% sample_n(1) 
    agent$x <- agent$B * agent$alpha
    agent$y <- agent$B * (1 - agent$alpha) / seller$p
    list("BUYER" = agent,
         "BOUGHT_FROM" = tibble(fromID = agent$ID, toID = seller$ID))
  }
}

## ~calcPrice~

## In the ~calcPrice~ transition function, the sellers summarise all the
## goods $x$ and $y$ they sold. As in their model state there are
## only the buyers which have created a link in the ~BOUGHT_FROM~
## network, they can do this by selecting the ~x~ and ~y~
## columns from the ~modelState\$BUYER~ tibble and then calculate
## the sum for those rows.

## The network ~BOUGHT_FROM~ is only constructed temporarily for
## this transition function, so no agent returns an edge of this network 
## and it will be also not added to the ~invariantTypes~, so this parameter
## will be ~c(BUYER,KNOWN\_SELLERS)~ to retain those agents and edges for the 
## next iteration of the simulation.

calcPrice <- function(agent, type, modelState) {
  if (type == "SELLER") {
    quant <- modelState$BUYER %>% select(x, y) %>% summarise_all(sum)
    agent$sum_y <- quant$y
    agent$p <- if (quant$x > 0) { quant$y / quant$x * agent$p } 
               else             { agent$p }
    list("SELLER" = agent)
  }
}

## For our model, we want to calculate the average price, but we
## can not just call ~sim$reduce("SELLER", "p", mean)~, as each seller
## has most certain sold a different amount of good $y$. We solve this
## by calculating $\frac{\sum{p * y}}{\sum y}$:

observeFunc <- function(current, sim) {
  current %>%
    bind_rows(
        tibble(
            sum_x = sim$reduce("BUYER", "x", sum),
            sum_y = sim$reduce("BUYER", "y", sum),
            p = sim$reduce("SELLER",
                           function(agent) { agent$p * agent$sum_y },
                           sum) / sim$reduce("SELLER", "sum_y", sum)
        ))
}

## So we have now every together to run and observe our simulation. We
## can then access the OBSERVER table via ~sim$global$OBSERVER~ and
## create a plot from this table, which shows the surplus demand for good $x$:

sim <- initSimulation(agentTypes = agentTypes,
                      edgeTypes = edgeTypes,
                      modelParams = modelParams,
                      stateInitFunc = initModel,
                      globalTypes = "OBSERVER",
                      globalInitFunc = initObserver)

for (i in 1:10) {
  applyTransition(sim, 
                  transitionFunction = calcDemand, 
                  network = "KNOWN_SELLERS", 
                  invariantTypes = c("SELLER","KNOWN_SELLERS"))
  applyTransition(sim, 
                  transitionFunction = calcPrice, 
                  network = "BOUGHT_FROM", 
                  invariantTypes = c("BUYER", "KNOWN_SELLERS"))
  updateGlobal(sim, "OBSERVER", observeFunc)
}

plot <- sim$global$OBSERVER %>% 
  rowid_to_column("step") %>% 
  ggplot(aes(x = step, y = sum_x - sum_y)) + geom_line() 

print(plot)
